dojo.provide("com.ibm.pb.control.NavigationController");

dojo.require("com.ibm.mm.builder.navigationHelper");
dojo.require("com.ibm.pb.iwidget.iWidgetBase");
dojo.require("com.ibm.pb.logging.Loggable");
dojo.require("com.ibm.pb.data.EnablerNavigationTreeModel");
dojo.require("com.ibm.pb.tree.DFSTraversal");
dojo.require("dojo.i18n");
dojo.requireLocalization("com.ibm.bundles", "NavWidget");
dojo.requireLocalization("com.ibm.bundles", "Theme");
dojo.requireLocalization("com.ibm.mm.builder", "coreWidgetsMessages");
dojo.require("dojo.hash");

/**
 * The NavigationController is meant to be a singleton object retrieved via the
 * com.ibm.pb.control.NavigationController.getController API.  It is responsible
 * for managing global navigation state and model information by providing access
 * to a dijit.tree.model implementation representing the navigation model.  It
 * also provides a selection model implementation that is mapped over top of the 
 * tree model and provides semantics as to the selection path from the currently
 * selected node in the navigation model up to the navigation root node.  These
 * constructs provide suitable models for backing interface widgets and for managing
 * public navigation selection events such as SWITCH_PAGE and PAGE_SWITCHED.  This
 * module provides logic for handling page switches both explicitly and via navigation
 * state changes.  It determines the correct context, updates the state, fires
 * the appropriate internal events to its selection model and tree model, and fires
 * the PAGE_SWITCHED event on successful selection changes.
 */
// not an iWidget, but uses the iWidget base for convenient provision of member variables
dojo.declare("com.ibm.pb.control.NavigationController", com.ibm.pb.iwidget.iWidgetBase, {
	
	navigationConstants: com.ibm.mashups.enabler.navigation.Constants,
	accFactory: com.ibm.mashups.enabler.model.state.AccessorFactory,
	treeModel: null,
	selectionModel: null,
	_changingSelection:false,
	currentThemeTemplate: "",

	// these fields are used for global page support
	lastNonGlobalPageId: null, // the last non-global page at any given time	
	lastSpaceId: null,  // the space being shown before the most recent update to the navigation
	globalPageLoadedSpaceId: null, // the space where global pages are opened		
	globalPagesOpened: [],   // the ids of the global pages currently showing
	globalPagesOpenedBy: [],   // the ids of the pages that opened each of the global pages currently showing

	constructor: function(){
		var lgr = this.LOGGER, mname = "constructor", me = this;
		lgr.entering(mname, arguments);
		this.themeBundle = dojo.i18n.getLocalization("com.ibm.bundles", "Theme");
		this.navWidgetBundle = dojo.i18n.getLocalization("com.ibm.bundles", "NavWidget");
		this.nm = com.ibm.mashups.enabler.navigation.Factory.getNavigationModel();
		this.nsm = com.ibm.mashups.enabler.model.state.NavigationStateModelFactory.getNavigationStateModel();
		this.spm = com.ibm.mashups.enabler.space.Factory.getSpaceModel();
		this.spaceAcc = this.accFactory.getSpaceAccessor(this.nsm);
		this.nm.addStrategy(new com.ibm.mashups.enabler.strategy.TreeLoadAheadStrategy(10, 1));	// 10 parents up, 1 level of children
		this.nm.addStrategy(new com.ibm.mashups.enabler.strategy.ListLoadAheadStrategy(50));	// 50 children at a time
		
		this.treeModel = new com.ibm.pb.data.EnablerNavigationTreeModel({nm: this.nm, nsm: this.nsm, spm: this.spm, delaySave: true});
		// Somewhat complex logic in getting the root, so keep it out of the tree model impl and just put it here
		this.treeModel.getRoot = dojo.hitch(this, "getRootNavigation");
		
		this.selectionModel = new com.ibm.data.SelectionModel({
			treeModel: this.treeModel,
			selectionPath: []
		});
		
		// Note: this logic was here for logging purposes but is throwing exceptions
		// in the case of global pages
//		var origChangeSelection = this.selectionModel.changeSelection;
//		this.selectionModel.changeSelection = dojo.partial(function(node){
//			lgr.trace("changeSelection", "Changing the selection model's selection ${0}", [node]);
//			return origChangeSelection.apply(this, arguments);
//		});

		// Restore global page information for cases of full page refresh
		this._restorePersistedGlobalPages();
		
		this._changingSelection = true;
		// calculate the current page asynchronously... will callback a selection model update
		this.bootstrapping = true;
		this._bootstrapState = false;
		this._calculateCurrentPage();
				
		this.evtSvc.subscribeEvent(this.iWidgetConstants.NAVSTATE_UPDATED, this, function(){
			var cbkName = "NavStateUpdatedHandler";
			lgr.entering(cbkName, arguments);
			// fetch the navigational state model again so it is not stale
			this.nsm = com.ibm.mashups.enabler.model.state.NavigationStateModelFactory.getNavigationStateModel();
			this.spaceAcc = this.accFactory.getSpaceAccessor(this.nsm);
			this.treeModel.nsm = this.nsm;
			
			var cid = this.getCurrentPageId();
			
			// Switch to the page if it is not the current page in navigation
			if(!this._changingSelection && cid !== this.currentPageId) {
				this.onSwitchPage(cid, {skipStateUpdate: true});
			}
			
			lgr.exiting(cbkName, arguments);
		});
		
		this.evtSvc.subscribeEvent(this.builderEvents.SWITCH_PAGE, this, "onSwitchPage");
		this.evtSvc.subscribeEvent(this.builderEvents.SWITCH_SPACE, this, "onSwitchSpace");
		this.evtSvc.subscribeEvent(this.builderEvents.REFRESH_NAVIGATION, this, "onRefreshNavigation");
		this.evtSvc.subscribeEvent(this.builderEvents.PAGE_SWITCHED, this, "_updateSelModel");
		this.evtSvc.subscribeEvent("com.ibm.mashups.builder.saveCheck", this, "dirtyPageCheck");

        dojo.subscribe("/dojo/hashchange", this, this._processHashChange);
		
		lgr.exiting(mname, arguments);
	},
	getSelectionModel: function() {
		return this.selectionModel;
	},
	getTreeModel: function() {
		return this.treeModel;
	},
	getThemeTemplate: function(node){
		return node.getMetaData("com.ibm.portal.theme.template.file.name.html") || "";
	},
	getThemeId: function(node) {
		var themeId = node.getThemeIdentifier ? node.getThemeIdentifier() : null;
		while(!themeId && node) {
			node = this.nm.getParent(node).start();
			themeId = (node && node.getThemeIdentifier) ? node.getThemeIdentifier() : null;
		}
		if(!themeId) {
			themeId = this.runtimeModel.getCurrentTheme();
		}
		return themeId.getID();
	},
	getThemeProfile: function(node) {
		var themeProfile = node.getMetaData("resourceaggregation.profile") || null;
		while(!themeProfile && node) {
			node = this.nm.getParent(node).start();
			if(node) {
				themeProfile = node.getMetaData("resourceaggregation.profile") || null;
			}
		}
		return themeProfile || "";
	},
	getRenderMode: function(node) {
		var renderType = node.getMetaData("com.ibm.portal.rendertype") || null;
		while(!renderType && node) {
			node = this.nm.getParent(node).start();
			if(node) {
				renderType = node.getMetaData("com.ibm.portal.rendertype") || null;
			}
		}
		return renderType || ibmConfig.defaultRenderingMode;
	},
	_visibleChildFinder: function(node){
		// Find the first visible child of the node
		var mname = "_visibleChildFinder";
		this._findFirstVisible(node, dojo.hitch(this, function(child){
			if(child && !(child instanceof Error)) {
				this._setCurrentNavNode(child);
			}
			else {
				this.LOGGER.severe(mname, "Error finding a visible child of navigation node ${0}: ${1}", [node, child]);
			}
		}));
	},
	_isNavLabel: function(model, node) {
		return node.getNavigationElementType() === this.navigationConstants.NAVIGATION_LABEL;
	},
	/**
	 * Sets the current navigation node of the controller internally, and updates any models
	 * and state information accordingly.
	 */
	_setCurrentNavNode: function(node, semantics) {
		// Set the current navigation node
		var mname = "_setCurrentNavNode", lgr = this.LOGGER, bootstrapping = this.bootstrapping;
		lgr.entering(mname, arguments);
		this.currentNavNode = node;
		this.currentPageId = node.getID();
		if(this.bootstrapping) {
			this.renderMode = this.getRenderMode(node);
			this.bootstrapping = false;
		}
		this.currentThemeTemplate = this.getThemeTemplate(node).toLowerCase();
		this.currentThemeProfile = this.getThemeProfile(node).toLowerCase();
		// Walk up the tree to load the ancestors asynchronously, if not already loaded
		// which will finally update the selection model at the end
		this._walkUpTree(node, dojo.delegate(semantics, {bootstrapping:bootstrapping}));
		this._updateNavigationState(this.currentPageId, semantics);
		if(this.renderMode === "iwidget") {
			// the title is set by theme markup or during initialization code;
			// this is meant primarily for client-side page changes
			this.setBrowserTitle(node);
		}
		lgr.exiting(mname, arguments);
	},
	_calculateCurrentPage: function() {
		var mname = "_calculateCurrentPage", lgr = this.LOGGER;
		lgr.entering(mname, arguments);
		var currentPageId = this.getCurrentPageId();
		if(!currentPageId) {
			this._bootstrapState = true;
			currentPageId = this.runtimeModel.getCurrentPage().getID();
		}
		else {
			this._bootstrapState = false;
			lgr.trace(mname, "Got current page from the navigation state", []);
		}
		if(!currentPageId) {
			// If we don't currently have a current page, then calculate it by loading the root
			// of the current navigation, which will load the root navigation node of the current
			// space we're in, or the default space if we're not in one.
			lgr.trace(mname, "No runtime model of current page, getting root navigation", []);
			this.treeModel.getRoot(dojo.hitch(this, "_visibleChildFinder"), dojo.hitch(this, function(err){
				lgr.severe(mname, "Error getting root navigation node ${0}", [err]);
			}));
		}
		else {
			lgr.trace(mname, "Found current page id ${0}", [currentPageId]);
			
			if (currentPageId && com.ibm.mm.builder.utils.spaceUtil.getSpaceIdByPageId(currentPageId)!=this.globalPageLoadedSpaceId
			  && !this._isOpenedGlobalPage(currentPageId)){
				// clear the global pages, as this initial page is in a different space from the one where the pages were opened
				this.onClearGlobalPage();
			}
			
			this.nm.find(currentPageId)
				.setFinishedCallback(dojo.hitch(this, function(node, status, params){
					if(status == 200) {
						this.refresh(node.getID());
					}
					else {
						// defect 16786: There will be page ID in the cookie, and such page won't be found in next install.
						// So in this case, we refresh the navigation with the nav root
						lgr.trace(mname, "Error finding the current page with id ${0} ${1}", [currentPageId, node||""]);
						var navRootID = this.nm.getRoot().start().getID();
						var spaceID = com.ibm.mm.builder.utils.spaceUtil.getSpaceIdByPageId(navRootID);
						com.ibm.mm.builder.utils.spaceUtil.setSpaceId(spaceID);
						this.refresh(navRootID);
					}
				}))
				.start(false);
		}
		lgr.exiting(mname, arguments);
	},
	_updateSelModel: function(){
		this.selectionModel.changeSelection(this.currentNavNode);
	},
	_walkUpTree: function(navNode, semantics) {
		// Recursively walks up the tree from the navNode until the root is reached
		// Finally internally updates the selection model with the current navigation node
		var mname = "_walkUpTree", lgr = this.LOGGER, me = this;
		if(navNode) {
			var dfd = this.nm.getParent(navNode).setFinishedCallback(function(node) {
				// walk up the tree asynchronously to guarantee all the ancestors are loaded before
				// firing the selection model update
				me._walkUpTree(node, semantics);
			});
			dfd.start(false);
		} 
		else {
			lgr.trace(mname, "Updating selection to ${0}", [this.currentNavNode.getID()]);
			if(semantics.bootstrapping) {
				// call this immediately during bootstrap since the PAGE_SWITCHED event isn't
				// guaranteed to be fired
				this._updateSelModel();
			}
			else {
				// we're listening to the PAGE_SWITCHED event and will call this._updateSelModel()
				// when that is fired
			}
		}
	},
	_findFirstVisible: function(parentNode, callback) {
		// Finds the first visible descendant of the given parentNode, and calls the callback
		// with that child navigation node as the argument
		var mname = "_findFirstVisible", lgr = this.LOGGER, me = this;
		lgr.entering(mname, arguments);
		var dft = new com.ibm.pb.tree.DFSTraversal(this.nm);
		dft.find(parentNode, function(model, navNode){
			return !me._isNavLabel(model, navNode) && !navNode.isHidden();
		}, function(foundNode){
			lgr.trace(mname, "Found visible non-label node ${0}, executing callback", [foundNode]);
			callback(foundNode);
		});
		lgr.exiting(mname, arguments);
	},
	getCurrentSpaceId: function() {
		var globalSpaceId = dojo.cookie(com.ibm.mm.builder.constant.cookieKey_globalPageOpenedSpaceId);
		if (globalSpaceId) {
			return globalSpaceId;
		} else {
			return this.spaceAcc.getSpaceID();
		}
	},
	getCurrentPageId: function() {
		var pageAcc = this.accFactory.getPageAccessor(this.nsm, this.spaceAcc.getSpaceID());
		return pageAcc.getPageID();
	},
	getRootNavigation: function(onItem, onError) {
		var mname = "getRootNavigation", lgr = this.LOGGER;
	    lgr.entering(mname, arguments);
		var id = this.getCurrentSpaceId(),
			params = {onItem:onItem, onError:onError};
		if(id) {
			var dfd = this.spm.find(id);
			dfd.setFinishedCallback(dojo.hitch(this, "_getRootBySpace"), params);
			dfd.start(false);
		}
		else {
			this._getRootBySpace(null, 200, params);
		}
		lgr.exiting(mname, arguments);
	},
	_getRootBySpace: function(space, status, params) {
		// This gets the root navigation node of a space, or gets the root navigation node
		// of the default space if the space argument is null.
		// Once retrieved, it calls the params.onItem function with the resulting node,
		// or the params.onError function if an error occurs at any point along the way.
		var mname = "_getRootBySpace", lgr = this.LOGGER;
		lgr.entering(mname, arguments);
		try {
			var navDfd;
			if(space) {
				navDfd = this.nm.find(space.getRootNavigation().getID());
			}
			else {
				navDfd = this.nm.getRoot();
			}
			navDfd.setFinishedCallback(dojo.hitch(this, function(navNode, statusCode){
				var cbkMname = mname + ":callback";
				lgr.entering(cbkMname, arguments);
				//lgr.trace(cbkMname, "Callback: ${0} ${1}", [navNode, statusCode]);
				
				if(navNode && !(navNode instanceof Error)) {
					this.root = navNode;
					if(params.onItem) {
						params.onItem(navNode);
					}
				}
				else if(params.onError){
					lgr.severe(cbkMname, "Error getting the root nav node in space ${0}: ${1}", [space, navNode]);
					params.onError(navNode);
				}
			}));
			lgr.trace(mname, "Finding root nav node in space ${0}", [space]);
			navDfd.start(false);
		} catch(err) {
			lgr.severe(mname, "Error starting find operation for the root nav node in space ${0}: ${1}", [space, err]);
		}
		lgr.exiting(mname, arguments);
	},
	/**
	 * Entry point for handling the SWITCH_SPACE event.
	 */
	onSwitchSpace: function(spaceId, semantics) {
		var mname = "onSwitchSpace", lgr = this.LOGGER, me = this;
		lgr.entering(mname, arguments);
		this.dirtyPageCheck(function(){
			me.switchSpaceHandler(spaceId, semantics);
		});
		lgr.exiting(mname, arguments);
	},
	switchSpaceHandler: function(spaceId, semantics) {
		var mname = "switchSpaceHandler", lgr = this.LOGGER, me = this;
		lgr.entering(mname, arguments);
		
		if (spaceId) {
			// we need to verify if the space is still in the model before we switch to it
			// in cases where the space has just been deleted and an event is sent to switch a space
			// or an invalid ID is used to send that event the enduser sees many JS exceptions
	        var spaceModel = com.ibm.mashups.enabler.space.Factory.getSpaceModel();
	    	var spaceNode = spaceModel.find(spaceId).start(true);
	    	if (!spaceNode) {
	    	    // this space node has been deleted
	    	    lgr.exiting(mname, arguments);
	        	return;
	    	}
		}

		// D18206
		// When switching spaces, check whether the currently opened page is a global page.
		// If so, need to update (fix) the pageID for the spaceID in the pageselection lookup index. 
		// At this point the pageselection for the current space will be referencing the global page.
		// If the user comes back to this space, the global page will be the one that is re-opened
		// (furthermore, the user will be taken to the space for the global page).
		// This fix here, "cleans up" the pageselection to refer back to the last non-global page
		// viewed in this space.
		if (this._isOpenedGlobalPage(this.currentPageId)) {
			this.accFactory.getPageAccessor(this.nsm, this.lastSpaceId).setPageID(this.lastNonGlobalPageId);				
		}

		if (spaceId != this.lastSpaceId){
			// clear global pages
			this.onClearGlobalPage();
		}				

		this.spaceAcc.setSpaceID(spaceId);	// set the new space so all subsequent calls will use that
		if (!spaceId) {
			dojo.cookie("inPersonalSpace", "true");
		}
		else {
			dojo.cookie("inPersonalSpace", null);
		}
		
		// get the current page from the nav state in the current space
		var pageId = this.getCurrentPageId(),
			pageNode = null;
		
		//switch to last page or first page.
		if(pageId) {
			pageNode = this.nm.find(pageId).start();
            // we need to update the navstate as setSpaceID will clear out the pageId from the navstate, so we have to set it back
            this.accFactory.getPageAccessor(this.nsm, spaceId).setPageID(pageId);
		}
		
		// if no page node was found matching that pageId
		if(!pageNode) {
			this.getRootNavigation(function(navNode){	// gets the root navigation of the current space
				me._findFirstVisible(navNode, function(firstVisibleNode){
					if(firstVisibleNode) {
						// broadcast SWITCH_PAGE to the page we've calculated
						me.evtSvc.broadcastEvent(me.builderEvents.SWITCH_PAGE, firstVisibleNode.getID());
					}
					else {
						// broadcast SWITCH_PAGE to the root page
						if(!spaceId) {
							com.ibm.mm.builder.utils.navigationUtil.autoAcceptPagesFromDefaultGroup();
						}
						me.evtSvc.broadcastEvent(me.builderEvents.SWITCH_PAGE, navNode.getID());
					}
				});
			}, function(err){
				lgr.severe(mname, "${0}", [err.message]);
			});
		}
		else {
			this.evtSvc.broadcastEvent(this.builderEvents.SWITCH_PAGE, pageId);
		}
		
		lgr.exiting(mname, arguments);
	},
	/**
	 * Entry point for handling the SWITCH_PAGE event.  The controller handles this event
	 * globally and singularly on behalf of all components that are interested in navigation
	 * selection changes.  This allows such logic to be consolidated and managed in one place.
	 * The semantics argument is optional and provides additional information which can be
	 * used by the controller during event handling.
	 */
	onSwitchPage: function(pageId, semantics){
		var mname = "onSwitchPage", lgr = this.LOGGER;
		lgr.entering(mname, arguments);

		// Mashup fix, do not refresh current page.
		if (pageId == this.currentPageId && !com.ibm.mm.builder.utils.pageUtil.isRoot(pageId)) {		
			if (this.spaceAcc.getSpaceID()!==null && this.spaceAcc.getSpaceID()!==undefined &&
			    this.lastSpaceId!==null && this.lastSpaceId!==undefined) {
				if (this.spaceAcc.getSpaceID()===this.lastSpaceId) {
					lgr.trace(mname, "skipping switchPage handling as already the current page");
					lgr.exiting(mname, arguments);
					return;
				}
			}
			else {
				lgr.trace(mname, "skipping switchPage handling as already the current page");
				lgr.exiting(mname, arguments);
				return;
			}			
		}
		// end fix

		// D18206
		// When switching a page, check if the page is a global page.
		// If it is not, remember it as the last non-global page opened.
		// The lastNonGlobalPageId variable is used in switchSpaceHandler() to set the last non-global
		// page visited in that space. 
		if (!this._isOpenedGlobalPage(pageId)) {
			this.lastNonGlobalPageId = pageId;
		}

		if (pageId && com.ibm.mm.builder.utils.spaceUtil.getSpaceIdByPageId(pageId)!=this.globalPageLoadedSpaceId
		  && !this._isOpenedGlobalPage(pageId)){
			// clear the global pages, as we've switched to a different space from the one where the pages were opened
			this.onClearGlobalPage();
		}

		this._changingSelection = true;
		
		this.nm.find(pageId).setFinishedCallback(dojo.hitch(this, "switchPageHandler"), semantics).start(false);
		
		this.lastSpaceId = this.getCurrentSpaceId();
		lgr.exiting(mname, arguments);
	},
	/**
	 * Handles the SWITCH_PAGE event to normalize the selection, potentially modifying it
	 * before updating internal models and firing the PAGE_SWITCHED event.  Takes
	 * the navigation node whose ID is equal to the payload of the SWITCH_PAGE event. 
	 */
	switchPageHandler: function(pageNode, status, semantics) {
		var mname = "switchPageHandler", lgr = this.LOGGER;
		lgr.entering(mname, arguments);
		if (this.renderMode != "iwidget") {
			// If we're in SSA mode, perform a full page refresh
			lgr.trace(mname, "Current rendertype: ${0} - performing a full page refresh", [this.renderMode]);
			this._reloadToPageWithConfirm(pageNode.getID());
		} /*** End SSA switch check ***/
		else {
			/*** Begin page node is label check ***/
			var isLabel = this._isNavLabel(this.nm, pageNode);
			lgr.trace(mname, "Is ${0} a label? ${1}", [pageNode, isLabel]);
			if(isLabel) {
				this._findFirstVisible(pageNode, dojo.hitch(this, function(node){
					if(node) {
						this.evtSvc.broadcastEvent(this.builderEvents.SWITCH_PAGE, node.getID());
					}
					else {
						lgr.trace(mname, "Could not find visible non-label descendant of ${0}, loading directly", [pageNode]);
						return this._postLabelCheck(pageNode, semantics);
					}
				}));
			} /*** End page node is label check ***/
			else {
				this._postLabelCheck(pageNode, semantics);
			}
		}	
		lgr.exiting(mname, arguments);
	},
	setLastNonGlobalPage: function() {
		var es = com.ibm.mashups.iwidget.services.ServiceManager.getService("eventService");
		es.broadcastEvent("com.ibm.mashups.builder.switchPage", this.lastNonGlobalPageId);
		
		this.onClearGlobalPage();
	},	
	clearLoginFlag: function() {
		dojo.cookie(com.ibm.mm.builder.constant.cookieKey_login_flag, null, {expires: -1});
	},	
	_postLabelCheck: function(pageNode, semantics) {
		var mname = "_postLabelCheck", lgr = this.LOGGER;
		lgr.entering(mname, arguments);
		/*** Begin SSA switch check ***/
		var renderType = this.getRenderMode(pageNode);
		if (renderType.toLowerCase() != "iwidget") {
			// If we are going into SSA mode, perform a full page refresh
			lgr.trace(mname, "Switched page rendertype: ${0}  - performing a full page refresh", [renderType]);
			this._reloadToPageWithConfirm(pageNode.getID());
		} /*** End SSA switch check ***/
		else {
			/*** Begin theme template check ***/
			var themeTemplate = this.getThemeTemplate(pageNode).toLowerCase();
			if (themeTemplate != this.currentThemeTemplate &&
				!((themeTemplate === "" && this.currentThemeTemplate === "theme.html") || 
				(themeTemplate === "theme.html" && this.currentThemeTemplate === ""))) {
				lgr.trace(mname, "Switched theme template from ${0} to ${1}", [this.currentThemeTemplate, themeTemplate]);
				this._reloadToPageWithConfirm(pageNode.getID());
			} /*** End theme template  check ***/
			else {
				/*** Begin theme profile check ***/
				var themeProfile = this.getThemeProfile(pageNode).toLowerCase();
				if (themeProfile != this.currentThemeProfile) {
					lgr.trace(mname, "Switched theme profile from ${0} to ${1}", [this.currentThemeProfile, themeProfile]);
					this._reloadToPageWithConfirm(pageNode.getID());
				} /*** End theme profile check ***/
				else {
					this.currentThemeTemplate = themeTemplate;
					this.currentThemeProfile = themeProfile;
					/*** Begin global page check ***/				
						this._postGlobalPageCheck(pageNode, semantics);
					/*** End global page check ***/
				}
			}
		}
		lgr.exiting(mname, arguments);
	},
	_postGlobalPageCheck: function(pageNode, semantics) {
		var mname = "_postGlobalPageCheck", lgr = this.LOGGER;
		lgr.entering(mname, arguments);
		var cb = dojo.hitch(this, function(){
			//this.isSwitching = true;
			try{
				this.refresh(pageNode.getID(), true, semantics);
			}catch(e){
				// if an error occurs during refresh, make sure we reset our _changingSelection
				// state
				this._changingSelection = false;
				lgr.severe(mname, "An error occured during page switching ${0}", [e]);
			}
			// refresh() will ensure that _changingSelection is reset appropriately when all
			// processing is completed
		});
		
		this.dirtyPageCheck(cb, cb);
		lgr.exiting(mname, arguments);
	},
	/**
	 * Sets the browser title using the title of the pageNode as part of the new title.
	 * This is called when a SWITCH_PAGE event proceeds passed all normalization
	 * steps to ensure that a proper selection is chosen.
	 * Extensions may override this to provide custom implementations.
	 */
	setBrowserTitle: function(pageNode) {
		var mname = "setBrowserTitle", lgr = this.LOGGER;
		lgr.entering(mname, arguments);

		var pageTitle = "";
		if(this.nm.getRoot().start().getID() != pageNode.getID()) {
			pageTitle = this.treeModel.getLabel(pageNode);
		}
		var trimmedTitle =  dojo.trim(pageTitle || ""),
			pageContext = com.ibm.mashups.enabler.context.Factory.getPageContext();
		
		// use the enabler API to set the HTML page title in browser
		if (pageContext) {
			var product = this.configSvc.getValue(this.configConstants.PRODUCT_NAME);
			if(product) {
				if (!pageTitle || trimmedTitle === "") {
					pageContext.setBrowserTitle(product);
				} 
				else {
					pageContext.setBrowserTitle(pageTitle);
				}
			}
			else if(pageTitle && trimmedTitle !== ""){
				pageContext.setBrowserTitle(pageTitle);
			}
		}
		lgr.exiting(mname, arguments);
	},
	
	dirtyPageCheck: function(callback, callbackNoSave, cancelcb){
		var mname = "dirtyPageCheck", lgr = this.LOGGER;
		lgr.entering(mname, arguments);
		if(com.ibm.pb.contextMenu.sharedActions.widgetEditModeChecker()) {
			// skip the page dirty check and display another confirm dialog if there is unsaved widget
			lgr.exiting(mname, arguments);
			return;
		}
		if(this.runtimeModel.getCurrentPage().isDirty()){
			lgr.trace(mname, "Page is dirty, display confirmation dialog", []);
			var cws = com.ibm.mm.builder.coreWidgetsStrings;
			/* Late lazy load */
			dojo["require"]("com.ibm.mm.builder.dialog.PageSaveConfirmDialog");
			var saveConfirmDlg = new com.ibm.mm.builder.dialog.PageSaveConfirmDialog({
		        contentWarning: cws.I_PAGE_SAVE_WARNING,
		        contentQuestion: cws.I_PAGE_SAVE_SUGGEST,
		        btnSaveText: cws.I_COMMON_SAVE,
		        btnNotSaveText: cws.I_PAGE_SAVE_NO_SAVE,
		        btnCancelText: cws.I_PAGE_SAVE_KEEP_EDIT,
		        hideNotSave: false,
		        callbackSave: callback,
		        callbackNotSave: callbackNoSave,
				callbackCancel: cancelcb
		    });
		    saveConfirmDlg.show();
		}else{
			lgr.trace(mname, "Page is NOT dirty, callback immediately", []);
			if(callback) callback();
		}
		lgr.exiting(mname, arguments);
	},
	_reloadToPageWithConfirm: function(pageId) {
		var cb = dojo.hitch(this, this.reloadToPage, pageId);
		this.dirtyPageCheck(cb);
	},
	reloadToPage: function(pageId) {
		var mname = "reloadToPage", lgr = this.LOGGER;
		lgr.entering(mname, arguments);
		var pageNode = this.nm.find(pageId).start();
		if(pageNode) {
			this._setCurrentNavNode(pageNode);
		}
		var redirectAddress = com.ibm.mashups.enabler.model.state.UrlGeneratorFactory.getUrlGenerator().getUrl(this.nsm, null, {skipState: true, pocURI: "nm:oid:"+pageId});
		lgr.trace(mname, "Navigating to URL immediately ${0}", [redirectAddress]);
		lgr.exiting(mname, arguments);
        
        // set the pageID cookie
        com.ibm.mm.builder.utils.navigationUtil.setPageIDCookie(pageId);

		var oldAddress = top.location.href;
		top.location.href = redirectAddress;
		if(this._removeHash(oldAddress) == this._removeHash(redirectAddress)) {
			setTimeout(function() {
                // this is a special marker with which certain extensions can behavior differently. In this particular case
                // it allows to write an unload extension that can differentiate from a normal browser close vs a refresh caused by mashups
                window.__page_reload_marker = true;
				// if we simply changed the hash, then setting the location before won't force a page refresh
				// so we force it here, attempting to minimize performance hit by getting from cache if possible
				top.location.reload(false);
			}, 50);
		}
	},
	_removeHash:function(s){
		var hash = s.indexOf("#");
		if(hash >= 0) {
			s = s.substring(0,hash);
		}
		return s;
	},
	onRefreshNavigation:function(){
		this.LOGGER.entering("onRefreshNavigation", arguments);
		try{
			this.refresh();
		}catch(e){
			var msg = dojo.string.substitute(com.ibm.mm.builder.coreWidgetsMessages.E_NAV_REFRESH_NAVIGATION_1,[e.message]);
			this.LOGGER.severe("onRefreshNavigation",msg);
			setTimeout(function(){
				throw e;	// the window can catch this and handle it outside our execution thread
			}, 1);
		}
		this.LOGGER.exiting("onRefreshNavigation", arguments);
	},
	/**
	 * Refreshes the navigation selection in the controller, potentially changing it
	 * depending on the value of the pid parameter.
	 */
	refresh: function(pid, isPageSwitch, semantics){
		var mname = "refresh", lgr = this.LOGGER, me = this, node, contentNode, spaceId = this.getCurrentSpaceId();
		lgr.entering(mname, arguments);
		if(pid) {
			// check if we have bootstrapped before and have a current page set already
			if(!this.bootstrapping) {
				/*** begin theme change check ***/
				lgr.trace(mname, "Checking for theme change");
				var currentThemeID = this.runtimeModel.getCurrentTheme().getID();
				var pageNode = this.nm.find(pid).start();
				if (pageNode && this.getThemeId(pageNode) != currentThemeID) {
					lgr.trace(mname, "Theme changed!");
					var cb = dojo.hitch(this, this.reloadToPage, pid);
					this.dirtyPageCheck(cb);
					lgr.exiting(mname, arguments);
					// don't set this._changingSelection to false because we're doing a full page refresh
					// and want it to stick until the browser reloads
					return;
				}
				/*** end theme change check ***/
			}
			else {
				// We're actually bootstrapping, so don't act like we're switching pages from an already selected page
				if(!semantics) {
					semantics = {};
				}
				semantics.skipStateUpdate = !this._bootstrapState;
				lgr.trace(mname, "Bootstrapping to current page ${0}", [pid]);
			}
		}
		else {
			pid = this.getCurrentPageId();
			if(!pid){
				lgr.trace(mname, "Find the first visible page from the current root");
				this.getRootNavigation(function(item){
					lgr.trace(mname, "Initiate page switch to ${0}", item);
					me.refresh(me.treeModel.getIdentity(item), isPageSwitch);
				}, function(err){
					this._changingSelection = false;
					lgr.severe(mname, "Error getting root navigation ${0}", err.message);
					setTimeout(function(){throw err;}, 1);
				});
				lgr.exiting(mname, arguments);
				return;
			}
		}
		// defect 16614
		// if the we are in nav root, we need to find whether there is child page visible
		// this page can be added from shares or hub
		var navRoot = this.nm.getRoot().start();
		if(pid == navRoot.getID()) {
            var interruptProcessing = false;
			this._findFirstVisible(navRoot, dojo.hitch(this, function(child){
				if(child && !(child instanceof Error)) {
                    interruptProcessing = true;
					this.onSwitchPage(child.getID());
					return;
				}
			}));
            if (interruptProcessing) {
                // we already found the correct page and initiated a page switch, no need to continue
                lgr.exiting(mname, arguments);
                return;
            }
		}
		
		// the nav must be reloaded on every refresh because deleting 
		// a page is handled the same as switching the page.  Since there
		// is no way to tell whether a page has been deleted we must
		// reload the nav to make sure the page is removed from the markup.
		// This will be fixed once Mashups created a deletePage iEvent.		
		if(pid){
			lgr.trace(mname, "Reloading navigation");
			node = this.nm.find(pid).start();
			if(node) {
				// node may have been deleted
				this._setCurrentNavNode(node, semantics);
			}
		}
		this._changingSelection = false;
		if(isPageSwitch) {
	        var nsm = com.ibm.mashups.enabler.model.state.NavigationStateModelFactory.getNavigationStateModel();        
			lgr.trace(mname, "Setting timeout to fire page switched event later");
			setTimeout(dojo.hitch(this, function(){
				lgr.trace(mname, "Sending page switched event");
				this.runtimeModel.getCurrentPage().setDirty(false);
				this.evtSvc.broadcastEvent(this.builderEvents.PAGE_SWITCHED, pid);
			}), 50);
		}
		lgr.exiting(mname, arguments);
	},
	/**
	 * The optional semantics arguments allows additional semantic information to be provided
	 * in paths where custom logic is required based on the context and circumstances.
	 * 
	 * The skipStateUpdate property specifies whether or not the state should be updated when
	 * the controller sets the current nav node.  The reason for this is because this function may
	 * be fired by listening to the SWITCH_PAGE event, in which case the navigation state
	 * should be updated by the controller.  Alternatively, this function may be called 
	 * on a path where the controller detects a NAVSTATE_UPDATED event separate
	 * from a SWITCH_PAGE event, but detects that the selection has changed as part of that
	 * state change.  In that case, the controller should not update the state since it
	 * has already been updated and the controller is merely synchronizing its internal
	 * models to reflect the new state.  If skipStateUpdate is true, then this does NOT
	 * update the navigation state.
	 */
	_updateNavigationState: function(pageId, semantics){
		var mname = "_updateNavigationState", lgr = this.LOGGER;
		lgr.entering(mname, arguments);
		this.runtimeModel.getCurrentPage().setID(pageId);
		if(!semantics || !semantics.skipStateUpdate) {
			// Do not update the navigation state if semantics.skipStateUpdate is truthy
			var spaceId = com.ibm.mm.builder.utils.spaceUtil.getSpaceIdByPageId(pageId);
			if (spaceId !== com.ibm.mm.builder.utils.spaceUtil.getCurrentSpaceId()) {
				com.ibm.mm.builder.utils.spaceUtil.setSpaceId(spaceId);
			}
			this.accFactory.getPageAccessor(this.nsm, spaceId).setPageID(pageId);
			this.navigationStateUpdated = true; // for testing in _processHashChange, to avoid firing page switch			
			this.nsm.commit({addToHistory: true}).start();
		}
		lgr.exiting(mname, arguments);
	},	
	
	onOpenPage: function(iEvent){
		var mname = "onOpenPage", lgr = this.LOGGER;
		lgr.entering(mname, arguments);
		// there are two kinds of payload format for ievent com.ibm.mashups.builder.openPage
		// 1. payload is a string : the page id, just switch to the page.
		// 2. payload is a JSON object : {"pid":"123456",eventsArray:{{"targetEvent": targetEvent1, "payload": payload1, "payloadType": payloadType1},{"targetEvent": targetEvent2, "payload": payload2, "payloadType": payloadType2}}}
		var openPageId = null;
		if (iEvent.payload.pid){
			openPageId = iEvent.payload.pid;	
		}else{
			openPageId = iEvent.payload;
		}
		var es = com.ibm.mashups.iwidget.services.ServiceManager.getService("eventService");
		es.broadcastEvent("com.ibm.mashups.builder.saveCheck", dojo.hitch(this,function(){
			var mname = "onOpenPage_saveCheck", lgr = this.LOGGER;
			lgr.entering(mname, arguments);
			if (this._isOpenedGlobalPage(openPageId)) {
				// if the global page is already opened, switch to it and reload it.
			}else{
				// otherwise this is a new global page, not currently open
				this.lastVisitedPage = com.ibm.mm.builder.utils.pageUtil.getCurrentPageId();
				if (this.lastSpaceId) {
					this.globalPageLoadedSpaceId = this.lastSpaceId;
				} else {
					this.globalPageLoadedSpaceId = this.getCurrentSpaceId();
				}
				this._addToGlobalPages(openPageId, this.lastVisitedPage);
				//create new tab in the tabbed navigation, and navigate to the new tab, with event broadcast to the page.
				var openPage = com.ibm.mm.builder.utils.pageUtil.getPageNode(openPageId);
				var globalSpaceRoot = com.ibm.mm.builder.utils.navigationUtil.getNavigationRootNode(this.globalPageLoadedSpaceId);
				var args = { 
					title:openPage.getTitle(dojo.locale),
					pageLoadedSpaceId:this.globalPageLoadedSpaceId,
					openPageId: openPageId
				};
				if (iEvent.payload.pid){
					args.eventsArray = iEvent.payload.eventsArray;
					args.lastVisitedPage = this.lastVisitedPage;
				}
				this.treeModel.newGlobalItem(args, globalSpaceRoot);
				lgr.trace(mname, "Broadcasting refreshNavigation event");
				es.broadcastEvent("com.ibm.mashups.builder.refreshNavigation");
			}
			setTimeout(dojo.hitch(this, function(){
				lgr.trace("onOpenPage_Timeout", "Broadcasting switch page event");
				if (iEvent.payload.eventsArray) {
					es.broadcastEvents(iEvent.payload.eventsArray, com.ibm.mm.builder.utils.pageUtil.getCurrentPageId(), openPageId);
				} else {
					es.broadcastEvent("com.ibm.mashups.builder.switchPage", openPageId);
				}
			}), 50);
			lgr.exiting(mname, arguments);
		}));	
		lgr.exiting(mname, arguments);
	},
	
	onClosePage: function(iEvent){
		var mname = "onClosePage", lgr = this.LOGGER;
		lgr.entering(mname, arguments);
		var closePageId = null;
		if (iEvent.payload) {
			if (iEvent.payload.closePageId) {
				closePageId = iEvent.payload.closePageId;
			} else if (!iEvent.payload.eventsArray) {
				closePageId = iEvent.payload;
			}
		}
		if (!closePageId) {
			closePageId = com.ibm.mm.builder.utils.pageUtil.getCurrentPageId();
		}
		if (!this._isOpenedGlobalPage(closePageId)){
			console.log("Error: you are closing a global page not opened now!");
			return;
		}
		var es = com.ibm.mashups.iwidget.services.ServiceManager.getService("eventService");
		es.broadcastEvent("com.ibm.mashups.builder.saveCheck", dojo.hitch(this,function(){
			var mname = "onClosePage_saveCheck", lgr = this.LOGGER;
			lgr.entering(mname, arguments);
			// go to last visited non-global page
			var openedPage = com.ibm.mm.builder.utils.pageUtil.getPageNode(closePageId);
			var args = {
				"lastVisitedPage": this._getGlobalPageOpenedBy(closePageId)
			};
			if (iEvent.payload && iEvent.payload.openPagePayload){
				args.openPagePayload = iEvent.payload.openPagePayload;
			}
			this._removeFromGlobalPages(closePageId);
			this.treeModel.deleteGlobalItem(args,openedPage);
			this.evtSvc.publishEvent(this.builderEvents.REFRESH_NAVIGATION);
			setTimeout(dojo.hitch(this, function(){
				lgr.trace("onClosePage_Timeout", "Broadcasting switch page event");
				if (args.openPagePayload){
					// do not switch to the last visited non-global page, we will open another global page instead.		
					es.broadcastEvent("com.ibm.mashups.builder.openPage", args.openPagePayload);
				}else{
					if (iEvent.payload && iEvent.payload.eventsArray) {
						es.broadcastEvents(iEvent.payload.eventsArray, null, args.lastVisitedPage, this.globalPageLoadedSpaceId);
					} else {
						es.broadcastEvent("com.ibm.mashups.builder.switchPage", args.lastVisitedPage);
					}
				}
			}), 50);	
			lgr.exiting(mname, arguments);
		}));
		lgr.exiting(mname, arguments);
	},
	
	onClearGlobalPage:function(){
		var mname = "onClearGlobalPage", lgr = this.LOGGER;
		lgr.entering(mname, arguments);
		for (var i in this.globalPagesOpened) {
			var openedPage = com.ibm.mm.builder.utils.pageUtil.getPageNode(this.globalPagesOpened[i]);
			var args = {"lastVisitedPage":this.lastVisitedPage};
			this.treeModel.deleteGlobalItem(args,openedPage);
		}
		this.globalPagesOpened = [];
		this.globalPagesOpenedBy = [];
		this.globalPageLoadedSpaceId = null;		
		dojo.cookie(com.ibm.mm.builder.constant.cookieKey_globalPageId, null, {expires: -1});
		dojo.cookie(com.ibm.mm.builder.constant.cookieKey_globalPageOpenedSpaceId, null, {expires: -1});
		dojo.cookie(com.ibm.mm.builder.constant.cookieKey_previous_pid, null, {expires: -1});
		lgr.exiting(mname, arguments);
	},
	
	_isOpenedGlobalPage: function(pageId){
		for (var i in this.globalPagesOpened) {
			if (this.globalPagesOpened[i] == pageId) {
				return true;
			}
		}
		return false;
	},
	
	_addToGlobalPages: function(pageId, openerPageId) {
		// find the index of the page to be deleted
		var pageIndex = -1;
		for (var i in this.globalPagesOpened) {
			if (this.globalPagesOpened[i] == pageId) {
				pageIndex = i;
			}
		}
		if (pageIndex >= 0) { // page already opened, just update info
			this.globalPagesOpened[pageIndex] = pageId;
			this.globalPagesOpenedBy[pageIndex] = openerPageId;
		} else { // new global page
			this.globalPagesOpened.push(pageId);
			this.globalPagesOpenedBy.push(openerPageId);
		}
		this._persistGlobalPages();
	},
	
	_removeFromGlobalPages: function(pageId){
		// find the index of the page to be deleted
		var pageIndex = -1;
		for (var i in this.globalPagesOpened) {
			if (this.globalPagesOpened[i] == pageId) {
				pageIndex = i;
			}
		}
		if (pageIndex >= 0) {
			// If there are any global pages that were opened by this global page, update
			// the opened by so they would return to the opener of this (deleted) global page
			for (var i in this.globalPagesOpened) {
				if (this.globalPagesOpenedBy[i] == pageId) {
					this.globalPagesOpenedBy[i] = this.globalPagesOpenedBy[pageIndex];
				}
			}
			// delete the page
			this.globalPagesOpened.splice(pageIndex, 1);
			this.globalPagesOpenedBy.splice(pageIndex, 1);
		}
		this._persistGlobalPages();
	},
	
	_getGlobalPageOpenedBy: function(pageId) {
		for (var i in this.globalPagesOpened) {
			if (this.globalPagesOpened[i] == pageId) {
				return this.globalPagesOpenedBy[i];
			}
		}
		return null;
	},
	
	_persistGlobalPages: function(){
		dojo.cookie(com.ibm.mm.builder.constant.cookieKey_globalPageId,dojo.toJson(this.globalPagesOpened));
		dojo.cookie(com.ibm.mm.builder.constant.cookieKey_previous_pid,dojo.toJson(this.globalPagesOpenedBy));
		dojo.cookie(com.ibm.mm.builder.constant.cookieKey_globalPageOpenedSpaceId,this.globalPageLoadedSpaceId);
	},
	
	_restorePersistedGlobalPages: function(){
		if (dojo.cookie(com.ibm.mm.builder.constant.cookieKey_globalPageOpenedSpaceId)) {
			this.globalPagesOpened = dojo.fromJson(dojo.cookie(com.ibm.mm.builder.constant.cookieKey_globalPageId));
			this.globalPagesOpenedBy = dojo.fromJson(dojo.cookie(com.ibm.mm.builder.constant.cookieKey_previous_pid));
			this.globalPageLoadedSpaceId = dojo.cookie(com.ibm.mm.builder.constant.cookieKey_globalPageOpenedSpaceId);
			for (var i in this.globalPagesOpened) {
				var openedPage = com.ibm.mm.builder.utils.pageUtil.getPageNode(this.globalPagesOpened[i]);
				var globalSpaceRoot = com.ibm.mm.builder.utils.navigationUtil.getNavigationRootNode(this.globalPageLoadedSpaceId);
				var args = { 
					title:openedPage.getTitle(dojo.locale),
					pageLoadedSpaceId:this.globalPageLoadedSpaceId,
					openPageId: this.globalPagesOpened[i]
				};
				this.treeModel.newGlobalItem(args, globalSpaceRoot);
			}
		}
	},
	
	_processHashChange: function(/*string*/hash) {
		var mname = "_processHashChange", lgr = this.LOGGER;
		lgr.entering(mname, arguments);
		if (!this.navigationStateUpdated) {
			var hashDec = decodeURIComponent(hash);
			var hashObj = dojo.queryToObject(hashDec);
			if (hashObj.pid) {
				var eventService = com.ibm.mashups.services.ServiceManager.getService("eventService");
				eventService.publishEvent(com.ibm.mashups.builder.iEvents.SWITCH_PAGE, hashObj.pid);
			}
		}
		this.navigationStateUpdated = false;
		lgr.exiting(mname, arguments);
	},

	onUnLoad: function(){
		var mname = "onUnLoad", lgr = this.LOGGER;
		lgr.entering(mname, arguments);
	    if (this.tabNavigation) {
			this.tabNavigation.destroy();
	        this.tabNavigation = null;
	    }
	    lgr.exiting(mname, arguments);
	}
	
});

(function(){
	var controller = null;
	com.ibm.pb.control.NavigationController.getController = function() {
		if(!controller) {
			controller = new com.ibm.pb.control.NavigationController();
		}
		return controller;
	};
})();
